"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
import re
from collections import deque
from synamic.core.contracts import BaseFsBackendContract
from .backends import FileSystemBackend
from synamic.core.contracts import HostContract, SiteContract, SynamicContract
from synamic.exceptions import SynamicInvalidCPathComponentError
from synamic.core.standalones.functions.sequence_ops import Sequence


regex_type = type(re.compile(""))
from synamic.core.standalones.functions.decorators import loaded, not_loaded


class _Patterns:
    path_sep = re.compile(r'[\\/]+')


class PathTree(object):
    def __init__(self, host):
        assert isinstance(host, HostContract)
        self.__host = host
        # TODO: assert os.path.isdir(host.site_root), "FileTree.__init__: _root must be a directory, `%s`"
        #  % host.site_root

        # default backend system
        self.__fs = FileSystemBackend()
        self.__is_loaded = False

    @classmethod
    def for_site(cls, site):
        assert isinstance(site, SiteContract)
        return cls(site)

    @property
    def host(self):
        return self.__host

    @property
    def is_site_host(self):
        return isinstance(self.__host, SiteContract)

    @property
    def is_synamic_host(self):
        return isinstance(self.__host, SynamicContract)

    @property
    def fs(self):
        return self.__fs

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    @classmethod
    def __str_path_to_comps(cls, path_str):
        # converting sting paths like ('x', 'a/b\\path_comp_str') to ('x', 'a', 'b', 'path_comp_str')
        # assert path_str.strip() != ''  # by empty path we mean site root
        str_comps = []
        for path_comp_str in _Patterns.path_sep.split(path_str):
            path_comp_str = path_comp_str.strip()
            str_comps.append(path_comp_str)
        return str_comps

    @classmethod
    def __sequence_path_to_comps(cls, path_sequence):
        str_comps = []
        for path_comp in path_sequence:
            if isinstance(path_comp, str):
                str_comps.extend(cls.__str_path_to_comps(path_comp))
            elif isinstance(path_comp, (list, tuple)):
                str_comps.extend(cls.__sequence_path_to_comps(path_comp))
            else:
                raise SynamicInvalidCPathComponentError(
                    f'Invalid component type for path: {type(path_comp)} whose value is {path_comp}'
                )

        return str_comps

    @classmethod
    def to_cpath_ccomps(cls, *path_comps):
        """Creates special cpath components that can have '' empty string on both ends
        To get components (without empty string on both ends) use path_comps property on cpath object"""
        comps = []
        for path_comp in path_comps:
            if isinstance(path_comp, str):
                comps.extend(cls.__str_path_to_comps(path_comp))
            elif isinstance(path_comp, (tuple, list)):
                comps.extend(cls.__sequence_path_to_comps(path_comp))
            elif type(path_comp) is cls.__CPath:
                # TODO: should check that called and caller are in the same site. This method is clcass method
                # and i cannot check the validity now.
                comps.extend(path_comp.path_comps)
            else:
                raise SynamicInvalidCPathComponentError(
                    f"Path comps must be list, tuple, string or __CPath object when it is not string: {type(path_comp)}"
                    f" where value is {path_comp}"
                )

        # remove empty '' part except for the first and last one
        _ = []
        for idx, comp in enumerate(comps):
            if idx in {0, len(comps) - 1}:
                # keep empty string for first and last part
                _.append(comp)
            else:
                # ignore the empty string.
                if comp != '':
                    _.append(comp)
        comps = _

        # Relative URL processing.
        _ = []
        for idx, comp in enumerate(comps):
            if comp == '..':
                if idx > 0:
                    del _[-1]
            elif comp == '.':
                # skip
                continue
            else:
                _.append(comp)
        else:
            if _ == []:
                _ = ['']
            if _[0] != '':
                _.insert(0, '')
        comps = _

        return tuple(comps)

    @classmethod
    def to_path_comps(cls, *path_comps):
        """The one returned by CPath.path_comps"""
        ccomps = cls.to_cpath_ccomps(*path_comps)
        # remove empty string from both ends
        if len(ccomps) > 1 and ccomps[-1] == '':
            ccomps = ccomps[:-1]
        if len(ccomps) > 1 and ccomps[0] == '':
            ccomps = ccomps[1:]

        comps = ccomps
        return comps

    def create_cpath(self, *path_comps, is_file=False, forgiving=False):
        """
        Create a Content Path object.
        :forgiving: if forgiving is True then it will not consider a result of path component ending in '' to be error.
            This method is only for end user who do not know deep details or for from template or when you want a relaxed
            way of doing things * Must not use in synamic core development.*
        """
        ccomps = self.to_cpath_ccomps(*path_comps)
        if not is_file:  # directory
            if ccomps[-1] != '':
                ccomps += ('', )
        else:  # file
            if not forgiving:  # is file & not forgiving
                assert ccomps[-1] != ''
            else:  # is file & forgiving.
                if len(ccomps) > 1 and ccomps[-1] == '':
                    ccomps = ccomps[:-1]

        path_obj = self.__CPath(self, self.__host, ccomps, is_file=is_file)
        return path_obj

    def create_file_cpath(self, *path_comps, forgiving=False):
        return self.create_cpath(*path_comps, is_file=True, forgiving=forgiving)

    def create_dir_cpath(self, *path_comps, forgiving=False):
        return self.create_cpath(*path_comps, is_file=False, forgiving=forgiving)

    def exists(self, *path) -> bool:
        comps = self.to_cpath_ccomps(*path)
        """Checks existence relative to the root"""
        return True if self.__fs.exists(self.__full_path__(comps)) else False

    def is_file(self, *path) -> bool:
        comps = self.to_cpath_ccomps(*path)
        fn = self.__full_path__(comps)
        return True if self.__fs.is_file(fn) else False

    def is_dir(self, *path) -> bool:
        comps = self.to_cpath_ccomps(*path)
        fn = self.__full_path__(comps)
        return True if self.__fs.is_dir(fn) else False

    def join(self, *content_paths):
        comps = self.to_cpath_ccomps(*content_paths)
        return self.create_cpath(comps)

    def open(self, file_path, *args, **kwargs):
        comps = self.to_cpath_ccomps(file_path)
        fn = self.__full_path__(comps)
        return open(fn, *args, **kwargs)

    def makedirs(self, *dir_path):
        comps = self.to_cpath_ccomps(*dir_path)
        full_p = self.__full_path__(comps)
        self.__fs.makedirs(full_p)

    @staticmethod
    def join_comps(*comps):
        """Comps must be strings
        Replacement for os path join as that discards empty string instead of putting a forward slash there"""
        _ = []
        for idx, comp in enumerate(comps):
            assert isinstance(comp, str), f'Provided type {type(comp)}'
            while comp.endswith(('/', '\\')):
                comp = comp[:-1]
            if idx > 0:
                while comp.startswith(('/', '\\')):
                    comp = comp[1:]

            if idx not in (0, len(comps) - 1):
                if comp == '':
                    # ignore empty string to avoid double slash in path
                    continue
            _.append(comp)
        return '/'.join(_)

    def get_full_path(self, *comps) -> str:
        comps = self.to_cpath_ccomps(*comps)
        return self.join_comps(self.__host.abs_root_path, *comps)

    def __full_path__(self, comps):
        # for internal use only where there is no normalization needed with self.to_cpath_comps
        """Comma separated arguments of path components or os.sep separated paths"""
        return self.join_comps(self.__host.abs_root_path, *comps)

    def __list_cpaths_loop2(self, starting_comps=(), files_only=None, directories_only=None, depth=None, exclude_cpaths=(), checker=None, respect_settings=True):
        return self.__ListCPathsLoop(
            self,
            starting_comps=starting_comps,
            files_only=files_only,
            directories_only=directories_only,
            depth=depth,
            exclude_cpaths=exclude_cpaths,
            checker=checker,
            respect_settings=respect_settings)()

    def list_cpaths(self, initial_path_comps=(), files_only=None, directories_only=None, depth=None, exclude_compss=(), checker=None, respect_settings=True):
        if type(initial_path_comps) is self.__CPath:
            assert initial_path_comps.is_dir
            starting_comps = initial_path_comps.path_comps
        else:
            starting_comps = self.to_cpath_ccomps(initial_path_comps)
        _exclude_compss = []
        for pc in exclude_compss:
            _exclude_compss.append(self.to_cpath_ccomps(pc))
        exclude_compss = tuple(_exclude_compss)

        dirs, files = self.__list_cpaths_loop2(starting_comps, files_only=files_only, directories_only=directories_only, depth=depth, exclude_cpaths=exclude_compss, checker=checker, respect_settings=respect_settings)
        return dirs, files

    def list_file_cpaths(self, initial_path_comps=(), depth=None, exclude_compss=(), checker=None, respect_settings=True):
        _, files = self.list_cpaths(initial_path_comps, files_only=True, depth=depth, exclude_compss=exclude_compss, checker=checker, respect_settings=respect_settings)
        return files

    def list_dir_cpaths(self, initial_path_comps='', depth=None, exclude_compss=(), checker=None, respect_settings=True):
        dirs, _ = self.list_cpaths(initial_path_comps, directories_only=True, depth=depth, exclude_compss=exclude_compss, checker=checker, respect_settings=respect_settings)
        return dirs

    def is_type_cpath(self, other):
        return type(other) is self.__CPath

    def __set_fs__(self, fs_instance):
        assert isinstance(fs_instance, BaseFsBackendContract)
        self.__fs = fs_instance

    class __ListCPathsLoop:
        def __init__(self, path_tree, starting_comps=(), files_only=None, directories_only=None, depth=None, exclude_cpaths=None, checker=None, respect_settings=True):
            self.path_tree = path_tree
            self.starting_comps = None
            self.files_only = files_only
            self.directories_only = directories_only
            self.depth = depth
            self.exclude_cpaths = None
            self.checker = checker
            self.respect_settings = respect_settings

            if starting_comps is None:
                self.starting_comps = ()
            else:
                self.starting_comps = self.path_tree.to_path_comps(starting_comps)

            for exclude_comps in exclude_cpaths:
                assert type(exclude_comps) is tuple, f"exclude_cpaths must contain tuple of strings as path." \
                                                     f" {exclude_comps} found"

            # check that files only and directories only both are not set to the Truth value
            if files_only is True:
                assert directories_only is not True
            if directories_only is True:
                assert files_only is not True

            # depth
            assert isinstance(depth, (type(None), int)), f"Type of depth must be None or int, {type(depth)}" \
                                                             f" found with value {depth}"
            if depth is None:
                self.depth = 2147483647

            # exclude cpaths validation
            _ = set()
            if exclude_cpaths is None:
                exclude_cpaths = set()
            for exclude_cpath in exclude_cpaths:
                assert self.path_tree.is_type_cpath(exclude_cpath)
                _.add(exclude_cpath)
            else:
                self.exclude_cpaths = _

            # default configs
            _dc = self.path_tree.host.system_settings['configs']
            self.__ignore_dirs_sw = _dc.get('ignore_dirs_sw', None)
            self.__ignore_files_sw = _dc.get('ignore_files_sw', None)

        def __call__(self, *args, **kwargs):
            """
                    A function to get all paths recursively starting from abs_root but returns a list of paths relative to the
                    .root
                    prefix_relative_root is fixed on every recursion
                    BUT next_relative_root isn't

                    exclude_comps_tuples: *components* list that are excluded from listing
                    checker: callables that accepts parameters: __ContentPath2 instance.
                    """
            absolute_root = self.path_tree.__full_path__(self.starting_comps)
            assert self.path_tree.fs.exists(absolute_root), f"Absolute root must exist: {absolute_root}"

            # new
            to_travel = deque([((*self.starting_comps, comp), 1) for comp in self.path_tree.fs.listdir(absolute_root)])
            directories = []
            files = []

            while len(to_travel) != 0:
                path_comps_n_depth = to_travel.popleft()
                path_comps = path_comps_n_depth[0]
                path_depth = path_comps_n_depth[1]
                if path_depth > self.depth:
                    break
                path_base = path_comps[-1]
                path_abs = self.path_tree.__full_path__(path_comps)

                if self.path_tree.fs.is_file(path_abs) and (self.files_only in (True, None)):
                    move_in = True
                    path_obj = self.path_tree.create_cpath(path_comps, is_file=True)
                    if self.checker is not None and not self.checker(path_obj):
                        move_in = False

                    elif self.respect_settings and path_base.startswith(self.__ignore_files_sw):
                        move_in = False

                    elif path_obj in self.exclude_cpaths:
                        move_in = False

                    if move_in:
                        files.append(path_obj)

                elif self.path_tree.fs.is_dir(path_abs) and (self.directories_only in (True, None)):
                    path_obj = self.path_tree.create_cpath(path_comps, is_file=False)
                    move_in = True
                    if self.checker is not None and not self.checker(path_obj):
                        move_in = False

                    elif self.respect_settings and path_base.startswith(self.__ignore_dirs_sw):
                        move_in = False

                    elif path_obj in self.exclude_cpaths:
                        move_in = False

                    if move_in:
                        directories.append(path_obj)
                        # Recurse
                        to_travel.extend(
                            tuple([((*path_comps, comp), path_depth + 1) for comp in self.path_tree.fs.listdir(path_abs)]))
                else:
                    raise Exception(f"ContentPath is neither dir, nor file: {path_abs}")
            return directories, files

    class __CPath:
        """
        CPath => Content Path
        Convention:
        1. Content Path will be indicated as cpath
        2. String Path will be indicated as path
        """
        def __init__(self, path_tree, site, cpath_special_comps, is_file=True):
            self.__path_tree = path_tree
            self.__cpath_special_comps = cpath_special_comps
            self.__site = site
            self.__is_file = is_file
            if is_file:
                assert cpath_special_comps[-1] != ''
            else:
                assert cpath_special_comps[-1] == ''

            # make path comps
            comps = self.__cpath_special_comps
            # remove empty string from both ends
            if len(comps) > 1 and comps[-1] == '':
                comps = comps[:-1]
            if len(comps) > 1 and comps[0] == '':
                comps = comps[1:]

            self.__path_comps = comps

        @property
        def site(self):
            return self.__site

        @property
        def parent_cpath(self):
            path_comps = self.path_comps
            if len(path_comps) <= 1:
                # contents/a-file
                # contents/dirA/b-file
                # * contents cannot have parent (==None) (Content Path is not for the site root or above dirs! SO -_-)
                # TODO: content should have prent with empty comps ()? If I put < 1 then app will stuck in infinite loop
                # Fix it.
                return None
            new_comps = path_comps[:-1]
            pp = self.__path_tree.create_cpath(new_comps, is_file=False)
            assert pp.is_dir, "Logical error, inform the developer (Sabuj)"  # Just to be safe from future bug.
            return pp

        @property
        def parent_cpaths(self) -> tuple:
            parent_paths = deque()
            pp = self.parent_cpath
            while pp is not None:
                parent_paths.appendleft(pp)
                pp = pp.parent_cpath

            return tuple(parent_paths)

        @property
        def relative_path(self):
            """
            Relative paths are relative from site root.
            """
            return self.__path_tree.join_comps(*self.__cpath_special_comps).replace('\\', '/')

        @property
        def path_comps(self):
            return tuple(self.__path_comps)

        @property
        def path_comps_w_site(self):
            return (*self.__site.id.components, *self.__path_comps[1:])

        @property
        def cpath_comps(self):
            """Special cpath comps that can have '' empty string on both ends - path tree's to components make this
            special components"""
            return tuple(self.__cpath_special_comps)

        @property
        def abs_path(self):
            return self.__path_tree.get_full_path(self.__cpath_special_comps)

        @property
        def is_file(self):
            return self.__is_file

        @property
        def is_dir(self):
            return not self.__is_file

        @property
        def basename(self):
            """
            Relative base name
            """
            return self.path_comps[-1]

        @property
        def basename_wo_ext(self):
            if self.is_file:
                base, dot, ext = self.basename.rpartition('.')
                if dot == '':
                    base = ext
                    ext = ''
            else:
                base = self.basename
            return base

        @property
        def dirname_comps(self):
            """
            Relative dirname
            """
            return self.__cpath_special_comps[:-1]

        @property
        def extension(self, dot_count=1):
            if self.is_file:
                dotted_parts = self.basename.rsplit(".", maxsplit=dot_count)
                if not len(dotted_parts) < dot_count + 1:
                    ext = ".".join(dotted_parts[-dot_count:])
                    return ext
            return ''

        def list_cpaths(self, files_only=None, depth=None, exclude_compss=(), checker=None, respect_settings=True):
            return self.__path_tree.list_cpaths(
                files_only=files_only,
                initial_path_comps=self,
                depth=depth,
                exclude_compss=exclude_compss,
                checker=checker,
                respect_settings=respect_settings
            )

        def list_files(self, depth=None, exclude_compss=(), checker=None, respect_settings=True):
            _, cfiles = self.list_cpaths(files_only=True, depth=depth, exclude_compss=exclude_compss, checker=checker, respect_settings=respect_settings)
            return cfiles

        def list_dirs(self, depth=None, exclude_compss=(), checker=None, respect_settings=True):
            dirs, _ = self.list_cpaths(files_only=False, depth=depth, exclude_compss=exclude_compss, checker=checker, respect_settings=respect_settings)
            return dirs

        def exists(self):
            """Real time checking"""
            return self.__path_tree.exists(self.__cpath_special_comps)

        def open(self, mode, *args, **kwargs):
            assert self.is_file, 'Cannot call open() on a directory: %s' % self.relative_path
            return self.__path_tree.open(self.__cpath_special_comps, mode, *args, **kwargs)

        def makedirs(self):
            assert self.is_dir
            return self.__path_tree.fs.makedirs(self.abs_path)

        def write_text(self, text):
            assert isinstance(text, str)
            assert self.is_file
            with self.open('w', encoding='utf-8') as fw:
                fw.write(text)

        def write_bytes(self, data):
            isinstance(data, (bytes, bytearray))
            assert isinstance(data, str)
            assert self.is_file
            with self.open('wb') as fw:
                fw.write(data)

        def write_stream(self, stream, close_on_done=False):
            assert hasattr(stream, 'read')
            assert self.is_file
            with self.open('wb') as fw:
                data = stream.read(1024)
                while data:
                    fw.write(data)
                    data = stream.read(1024)
            if close_on_done:
                stream.close()

        def write_text_stream(self, stream, close_on_done=False):
            assert hasattr(stream, 'read')
            assert self.is_file
            with self.open('w', encoding='utf-8') as fw:
                data = stream.read(1024)
                while data:
                    fw.write(data)
                    data = stream.read(1024)
            if close_on_done:
                stream.close()

        def make_file(self):
            assert self.is_file
            with self.open('wb') as fw:
                pass

        def join(self, *path_str_or_cmps, is_file=True, forgiving=False):
            """Creates a new path joining to this one"""
            comps = self.__path_tree.to_cpath_ccomps(*path_str_or_cmps) # [p for p in re.split(r'[\\/]+', path_str) if p != '']
            if self.is_dir:
                new_path_comps = (*self.path_comps, *comps)
            else:
                new_path_comps = (*self.path_comps[:-1], self.path_comps[-1] + comps[0], *comps[1:])
            return self.__path_tree.create_cpath(new_path_comps, is_file=is_file, forgiving=forgiving)

        @staticmethod
        def __process_regex(regex, ignorecase=True):
            """Matches against relative path"""
            if isinstance(regex, str):
                if ignorecase:
                    regex = re.compile(regex, re.IGNORECASE)
                else:
                    regex = re.compile(regex, re.IGNORECASE)
            else:
                assert type(regex) is regex_type, "regex argument must provide compiled regular expression or string"
            return regex

        def match(self, regex, ignorecase=True):
            """Matches against relative path"""
            regex = self.__process_regex(regex, ignorecase)
            return regex.match(self.relative_path)

        def match_basename(self, regex, ignorecase=True):
            """"""
            regex = self.__process_regex(regex, ignorecase)
            return regex.match(self.basename)

        def match_extension(self, regex, ignorecase=True):
            """"""
            regex = self.__process_regex(regex, ignorecase)
            return regex.match(self.extension)

        def startswith(self, *comps):
            ccomps = self.__path_tree.to_cpath_ccomps(comps)
            if not (len(self.path_comps) < len(ccomps)):
                if self.path_comps[:len(ccomps)] == ccomps:
                    return True
            return False

        def endswith(self, *comps):
            ccomps = self.__path_tree.to_cpath_ccomps(comps)
            if not (len(self.path_comps) < len(ccomps)):
                if self.path_comps[-len(ccomps):] == ccomps:
                    return True
            return False

        def getmtime(self):
            return self.__path_tree.fs.getmtime(self.abs_path)

        def getctime(self):
            return self.__path_tree.fs.getctime(self.abs_path)

        @property
        def id(self):
            return '/'.join(self.path_comps)

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return False
            return self.path_comps == other.path_comps and self.is_file == other.is_file

        def __hash__(self):
            return hash(self.path_comps)

        def __str__(self):
            return f"CPath: {self.relative_path}"

        def __repr__(self):
            return repr(str(self))



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.rst', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='synamic',
    version='0.5.8.dev5',
    description='Synamic',
    author='Md. Sabuj Sarker',
    author_email='md.sabuj.sarker@gmail.com',
    url='https://github.com/SabujXi/Synamic',
    packages=find_packages('src'),
    # packages=find_packages('src', exclude=('synamic',)),
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': [
            'synamic = synamic.entry_points.synamic_shell:main'
        ]
    },
    data_files=[('', ['README.rst', 'LICENSE'])],
    package_data={
        '': ['*.model', '*.syd', '*.txt', '*.rst', '*.md', '*.markdown'],
    },
    install_requires=[
        'Jinja2>=2.10',
        'MarkupSafe>=1.0',
        'mistune>=0.8.3',
        'Pillow>=5.2.0',
        'sly>=0.3',
        'aiohttp>=3.4.4',
    ],
    python_requires='>=3.6',
    long_description=long_description
)

# print(find_packages('src', exclude=('synamic',)))



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


def _callable_when_not_finalized(meth):
    def wrapper(self, *args, **kwargs):
        if self.finalized:
            raise Exception("FinalizableDict cannot be modified once it is finalized")
        return meth(self, *args, **kwargs)
    return wrapper


class FinalizableDict(dict):
    __slots__ = ('__finalized',)

    def __init__(self, *itable, **kwargs):
        super().__init__(*itable, **kwargs)
        self.__finalized = False

    @property
    def finalized(self):
        return self.__finalized

    @_callable_when_not_finalized
    def finalize(self):
        self.__finalized = True

    @_callable_when_not_finalized
    def __setitem__(self, key, value):
        return super().__setitem__(key, value)

    @_callable_when_not_finalized
    def __delitem__(self, key):
        return super().__delitem__(key)

    @_callable_when_not_finalized
    def update(self, m, **kwargs):
        return super().update(m, **kwargs)

    @_callable_when_not_finalized
    def clear(self):
        return super().clear()



import io
from synamic.core.contracts import ContentContract, CDocType
from synamic.core.services.content.toc import Toc


class MarkedContent(ContentContract):
    def __init__(self, site, body_text, cfields):
        self.__site = site
        self.__body_text = body_text
        self.__cfields = cfields
        self.__cmodel = cfields.cmodel

        # validation
        assert CDocType.is_text(self.__cfields.cdoctype)

        self.__toc = None
        self.__body = None

    @property
    def site(self):
        return self.__site

    def get_stream(self):
        template_name = self.__cfields.get('template', 'default.html')
        templates = self.__site.get_service('templates')
        res = templates.render(template_name, context={
            'site': self.__site,
            'content': self
        })
        f = io.BytesIO(res.encode('utf-8'))
        return f

    def __render_body(self):
        body = self.__body
        if body is None:
            toc = Toc()
            # try with model converter
            body_field_key = self.__cmodel.body_field
            if body_field_key is None:
                body_field_key = '__body__'

            model_field = self.__cmodel.get(body_field_key)
            if model_field is not None:
                body = model_field.converter(self.__body_text, value_pack={
                    'toc': toc,
                    'from_cpath': self.cpath
                })

            # try with markdown renderer
            else:
                markdown_renderer = self.__site.get_service('types').get_converter('markdown')

                body = markdown_renderer(self.__body_text, value_pack={
                    'toc': toc,
                    'from_cpath': self.cpath
                })
            self.__toc = toc
            self.__body = body

    @property
    def body(self):
        self.__render_body()
        return self.__body

    @property
    def body_as_string(self):
        return self.body

    @property
    def body_as_bytes(self):
        return self.body.encode('utf-8')

    @property
    def cfields(self):
        return self.__cfields

    @property
    def toc(self):
        self.__render_body()
        return self.__toc

    def __getitem__(self, key):
        return self.__cfields[key]

    def __getattr__(self, key):
        return getattr(self.__cfields, key)

    def __str__(self):
        return "<%s>\n%s\n" % (self.cfields.cpath.relative_path, self['title'])

    def __repr__(self):
        return str(self)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
import abc
import enum


@enum.unique
class CDocType(enum.Enum):
    # later, intending the use of auto() - currently this project in 3.5 and auto() is available in 3.6
    BINARY_DOCUMENT = "BINARY_DOCUMENT"
    GENERATED_BINARY_DOCUMENT = "GENERATED_BINARY_DOCUMENT"
    TEXT_DOCUMENT = "TEXT_DOCUMENT"
    GENERATED_TEXT_DOCUMENT = "GENERATED_TEXT_DOCUMENT"
    HTML_DOCUMENT = "HTML_DOCUMENT"
    GENERATED_HTML_DOCUMENT = "GENERATED_HTML_DOCUMENT"

    META_DOCUMENT = "META_DOCUMENT"
    NOURL_DOCUMENT = "NOURL_DOCUMENT"
    DIRECTORY = "DIRECTORY"
    UNSPECIFIED = "UNSPECIFIED"

    @classmethod
    def is_binary(cls, other, not_generated=False):
        if not_generated:
            return other in (cls.BINARY_DOCUMENT, )
        else:
            return other in (cls.BINARY_DOCUMENT, cls.GENERATED_BINARY_DOCUMENT)

    @classmethod
    def is_text(cls, other, not_generated=False):
        if not_generated:
            return other in (cls.TEXT_DOCUMENT, cls.HTML_DOCUMENT, )
        else:
            return other in (cls.TEXT_DOCUMENT, cls.GENERATED_TEXT_DOCUMENT, cls.HTML_DOCUMENT, cls.GENERATED_HTML_DOCUMENT)

    @classmethod
    def is_html(cls, other, not_generated=False):
        if not_generated:
            return other in (cls.HTML_DOCUMENT, )
        else:
            return other in (cls.HTML_DOCUMENT, cls.GENERATED_HTML_DOCUMENT)

    @classmethod
    def is_file(cls, other, not_generated=False):
        return cls.is_text(other, not_generated=not_generated) or cls.is_binary(other, not_generated=not_generated)

    @classmethod
    def is_generated(cls, other):
        return other in (cls.GENERATED_HTML_DOCUMENT, cls.GENERATED_TEXT_DOCUMENT, cls.GENERATED_BINARY_DOCUMENT)


class ContentContract(metaclass=abc.ABCMeta):

    __cdoc_types = CDocType

    @property
    @abc.abstractmethod
    def site(self):
        pass

    @property
    @abc.abstractmethod
    def cfields(self):
        pass

    @abc.abstractmethod
    def get_stream(self):
        """
        This will be a file like object. 
        """

    @property
    @abc.abstractmethod
    def body(self):
        pass

    @property
    @abc.abstractmethod
    def body_as_bytes(self):
        pass

    @property
    @abc.abstractmethod
    def body_as_string(self):
        pass

    def __str__(self):
        return "Content with url -> %s" % self.cfields.curl.path_as_str_w_site

    def __repr__(self):
        return repr(self.__str__())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.cfields.curl == other.cfields.curl

    def __hash__(self):
        return hash(self.cfields.curl)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import re
import mistune
schemed_url_pat = re.compile(r'^[a-zA-Z0-9_]+://')


class SynamicBaseRenderer(mistune.Renderer):
    def __init__(self, site, value_pack=None, **kwargs):
        super().__init__(**kwargs)
        self.__site__ = site
        self.__value_pack__ = value_pack if value_pack is not None else {}

    def image(self, src, title, alt_text):
        if not title:
            title = ''
        if not alt_text:
            alt_text = title
        url = self.__site__.object_manager.getc(src, relative_cpath=self.__value_pack__.get('from_cpath', None))
        return f"<img src='{url}' title='{title}' alt='{alt_text}' class='img-responsive center-block'>"

    def header(self, text, level, raw=None):
        html_id = self.__text_2_html5_id(text)
        toc = self.__value_pack__.get('toc', None)
        if toc is not None:
            toc.add(level, text, html_id)
        return f"<h{level} id='{html_id}'>{text}</h{level}>"

    def __text_2_html5_id(self, text):
        rpl_spc = re.sub(r'\s', '_', text)
        final = re.sub(r'["\']', '-', rpl_spc)
        return final


class SynamicContentRenderer(SynamicBaseRenderer):
    def link(self, link, title, content):
        url = self.__site__.object_manager.getc(link, relative_cpath=self.__value_pack__.get('from_cpath', None))
        # TODO: process link if a file is referenced directly.
        # TODO: Assert if it returns a str.
        return f"<a href='{url}' title='{title}'>{content}</a>"


class SynamicChaptersRenderer(SynamicContentRenderer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def link(self, link, title, content):
        if schemed_url_pat.match(link):
            url_str = super().link(link, title, content)
        else:
            curl = self.__site__.object_manager.get_curl_by_filename(
                link,
                relative_cpath=self.__value_pack__.get('from_cpath', None)
            )
            url_str = curl.url
            # process the link
            curl_bucket = self.__value_pack__.get('curl_bucket', None)
            if curl_bucket is not None:
                curl_bucket.append(curl)
        return f"<a href='{url_str}' title='{title}'>{content}</a>"


def render_content_markdown(site, text, value_pack=None):
    renderer = SynamicContentRenderer(site, value_pack)
    md = mistune.Markdown(renderer=renderer)
    return md.render(text)


def render_chapters_markdown(site, text, value_pack=None):
    renderer = SynamicChaptersRenderer(site, value_pack)
    md = mistune.Markdown(renderer=renderer)
    return md.render(text)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
from urllib.parse import urlparse, urlunparse
from synamic.core.synamic.router.url import ContentUrl


class _SiteSettings:
    def __init__(self, site, syd):
        self.__site = site
        self.__syd = syd
        self.__site_address = None
        self.__site_base_path = None

    @property
    def site_address(self):
        if self.__site_address is None:
            # calculate dev server address when dev server running.
            dev_site_address = self.__site.synamic.dev_params.get('site_address', None)
            if dev_site_address:
                site_address = dev_site_address
            # calculate from settings/configs
            else:
                site_address = self.__syd['site_address']

            site_url_struct = urlparse(site_address)
            scheme = site_url_struct.scheme
            netloc = site_url_struct.netloc
            path = '/'.join(
                ContentUrl.path_to_components(
                    site_url_struct.path,
                    self.__site.id.components
                )
            )
            site_url = urlunparse([scheme, netloc, path, '', '', ''])
            site_address = site_url

            if not site_address.endswith('/'):
                site_address += '/'

            # setting site address.
            self.__site_address = site_address
            # setting site base path.
            # if not path.endswith('/'):
            #     path += '/'
            self.__site_base_path = path

        return self.__site_address

    @property
    def site_base_path(self):
        if self.__site_base_path is None:
            _ = self.site_address
        return self.__site_base_path

    @property
    def origin_syd(self):
        return self.__syd

    def get(self, dotted_key, default=None):
        value = self.__syd.get(dotted_key, default=default)
        if value == default:
            # try in system
            value = self.__site.system_settings.get(dotted_key, default)
        return value

    def __getitem__(self, item):
        value = self.get(item, None)
        if value is None:
            raise KeyError('Key `%s` in settings was not found' % str(item))
        else:
            return value

    def __getattr__(self, key):
        value = self.get(key, None)
        if value is None:
            raise AttributeError('Attribute `%s` in settings was not found' % str(key))
        else:
            return value

    def __contains__(self, item):
        res = self.get(item, None)
        if res is None:
            return False
        else:
            return True

    def __str__(self):
        return str(self.__)

    def __repr__(self):
        return repr(self.__str__())


class SiteSettingsService:
    def __init__(self, site):
        self.__site = site
        self.__is_loaded = False

    def load(self):
        self.__is_loaded = True

    def make_site_settings(self):
        system_settings = self.__site.system_settings

        # all parent settings are merged
        parent_settings_syd = []
        site = self.__site
        while site.has_parent:
            parent_settings_syd.append(
                self.__site.parent.settings.origin_syd
            )
            site = site.parent
        parent_settings_syd.reverse()

        # site specific settings and private settings.
        site_om = self.__site.object_manager
        site_settings_cfile = self.__site.path_tree.create_file_cpath('settings.syd')
        site_settings_private_cfile = self.__site.path_tree.create_file_cpath('settings.private.syd')
        if site_settings_cfile.exists():
            parent_settings_syd.append(
                site_om.get_syd(site_settings_cfile)
            )
        if site_settings_private_cfile.exists():
            parent_settings_syd.append(
                site_om.get_syd(site_settings_private_cfile)
            )

        # construct final syd
        new_syd = system_settings.new(*parent_settings_syd)
        # construct site settings object
        site_settings = _SiteSettings(self.__site, new_syd)
        return site_settings



import sly
from sly import Lexer, Parser
from collections import namedtuple
from synamic.exceptions import SynamicGetCParsingError
from synamic.core.object_manager.query import generate_error_message

UrlStruct = namedtuple('UrlStruct', ('scheme', 'site_id', 'keys', 'path'))


def parse_getc(text) -> UrlStruct:
    """
    """
    text = text.strip()
    parser = GetCParser(text)
    lexer = GetCLexer()
    try:
        res = parser.parse(lexer.tokenize(text))
    except sly.lex.LexError as e:
        text_rest = e.text
        err_txt = generate_error_message(text, text_rest)

        raise SynamicGetCParsingError(
            f'Lexical error at index: {e.error_index}\nDetails: {err_txt}'
        )
    else:
        return res


class GetCLexer(Lexer):
    tokens = {'SCHEME_NAME', 'SCHEME_SEP', 'SITE_ID', 'KEY', 'KEY_SEP', 'PATH'}
    PATH = r'[^:]+$'
    SCHEME_NAME = r'[a-zA-Z]+(?=://)'
    SCHEME_SEP = r'://'
    SITE_ID = r'[^\s:/\\]+(~[^\s:/\\]+)+|~[^\s:/\\]+|~'
    # sync ~ this value with system system_settings['configs.site_id_sep']
    KEY = r'[a-zA-Z_][a-zA-Z0-9_-]*'
    KEY_SEP = r':'


class GetCParser(Parser):
    def __init__(self, text):
        self.__text = text

    # Get the token list from the lexer (required)
    tokens = GetCLexer.tokens

    # Grammar rules and actions
    @_('PATH',  # normal relative url
       'SCHEME_NAME SCHEME_SEP PATH',  # normal absolute url
       'SCHEME_NAME SCHEME_SEP keys PATH',  # geturl
       'SCHEME_NAME SCHEME_SEP SITE_ID KEY_SEP keys PATH'  # geturl
       )
    def query(self, p):
        length = len(p)
        if length == 1:
            struct = UrlStruct(scheme='', keys=tuple(), site_id=None, path=p[0])
        elif length == 3:
            struct = UrlStruct(scheme=p[0], keys=tuple(), site_id=None, path=p[2])
        elif length == 4:
            struct = UrlStruct(scheme=p[0], keys=p[2], site_id=None, path=p[3])
        else:
            assert length == 6
            struct = UrlStruct(scheme=p[0], keys=p[4], site_id=p[2], path=p[5])
        return struct

    @_('KEY KEY_SEP',
       'KEY KEY_SEP keys')
    def keys(self, p):
        if len(p) == 2:
            return p[0],
        else:
            assert len(p) == 3
            return p[0], p[2][0]

    def error(self, p):
        text = self.__text
        if not p:
            text_rest = ''
            err_txt = generate_error_message(text, text_rest)
            err = SynamicGetCParsingError(
                f"GetC End of param text before tokens could be parsed sensibly.\nDetails: {err_txt}"
            )
        else:
            text_rest = self.__text[p.index:]
            err_txt = generate_error_message(text, text_rest)
            err = SynamicGetCParsingError(
                f"GetC param Parsing error at token: {p.type} -> {p.value}."
                f" Error at index {p.index}.\nDetails:{err_txt}"
            )
        raise err


if __name__ == '__main__':
    while True:
        try:
            text = input('parse > ')
        except EOFError:
            raise
        if text:
            try:
                for token in GetCLexer().tokenize(text):
                    print(token)
                res = parse_getc(text)
            except SynamicGetCParsingError as e:
                print(e.message)
            else:
                print(res)



from synamic.core.services.event_system.event_types import EventTypes
from synamic.exceptions.synamic_exceptions import LogicalError


def _not_completed_decorator(method):
    def method_wrapper(self, *args, **kwargs):
        if self.completed:
            raise LogicalError("This method (%s) can only be used when the event was not completed" % method.__name__)
        return method(self, *args, **kwargs)
    return method_wrapper


class Event:
    def __init__(self, subject, **kwargs):
        self.__subject = subject
        self.__completed = False
        self.__container = {}
        self.__container.update(kwargs)

    @property
    @_not_completed_decorator
    def subject(self):
        return self.__subject

    @property
    def completed(self):
        return self.__completed

    @_not_completed_decorator
    def mark_complete(self):
        self.__completed = True
        del self.__subject
        self.__container.clear()
        del self.__container

    @_not_completed_decorator
    def get(self, key, *args, **kwargs):
        return self.__container.get(key, *args, **kwargs)

    @_not_completed_decorator
    def keys(self):
        return self.__container.keys()

    @_not_completed_decorator
    def values(self):
        return self.__container.values()

    @_not_completed_decorator
    def items(self):
        return self.items()

    @_not_completed_decorator
    def __getitem__(self, item):
        return self.__container.__getitem__(item)

    @_not_completed_decorator
    def __setitem__(self, key, value):
        return self.__container.__setattr__(key, value)

    @_not_completed_decorator
    def __iter__(self):
        return self.__container.__iter__()

    @_not_completed_decorator
    def __contains__(self, item):
        return self.__container.__contains__(item)


class Handler:
    def __init__(self, handler_callable):
        assert callable(handler_callable)
        self.__handler_callable = handler_callable

    def __call__(self, event, **kwargs):
        assert type(event) is Event
        return self.__handler_callable(event, **kwargs)


# DONE: Event system must not be class based, instead an event system must be instantiated as we may have
# multiple copy of synamic in the same process.
# also, it is needed for synamic.reload()

class EventSystem:
    """
    Handler config:
    handler(what)
    e.g.:
        what: site obj, content service/module object, etc.
        
    Custom event handlers are primarily for external plugins. So, anyone is can trigger that.
    BUT, system events (EventTypes) are only for site and only site can trigger that.
    So, as a protection only the first time the trigger method can be achieved and any subsequent call will raise
    exception.
    
    * System event __document_types can be derived from the EventTypes enum, but custom event __document_types are specified by string.
    The strings are lowercased to avoid any type of probable conflict from inexperienced users/plugin-developers.
    
    Handler functions will be called with an event object. An event object will hold a subject,
    and a map of other objects where strings are used as keys.
    
    """

    # a site/config cannot possess or create more than one event system.
    __instances_map = {}

    @classmethod
    def get_event_system(cls, site):
        # return cls.__instances_map.get(site, None)
        raise NotImplemented

    def __init__(self, site_or_synamic):
        assert site_or_synamic not in self.__instances_map
        self.__instances_map[site_or_synamic] = self
        #  event maps: key => event type, value => ordered dict of event handlers
        self.__get_trigger_count = 0
        self.__event_map = {}
        self.__custom_event_map = {}

        # will use the following when an int will be used to cancel a registered event
        self.__last_event_id = 0
        self.__last_custom_event_id = 0

        for et in EventTypes:
            self.__event_map[et] = []

        self.__is_loaded = False

    def load(self):
        self.__is_loaded = True

    def add_event_handler(self, etype: EventTypes, handler):
        assert type(etype) is EventTypes
        assert type(handler) is Handler
        earr = self.__event_map[etype]
        earr.append(handler)

    def add_custom_event_handler(self, ename, handler):
        assert type(ename) is str
        assert type(handler) is Handler
        ename = ename.lower()
        earr = self.__event_map.get(ename, None)
        if earr is None:
            earr = self.__event_map[ename] = []
        earr.append(handler)

    def remove_event_handler(self, etype, handler):
        assert type(etype) is EventTypes
        assert type(handler) is Handler
        earr = self.__event_map[etype]
        earr.remove(handler)

    def remove_custom_event_handler(self, ename, handler):
        assert type(ename) is str
        assert type(handler) is Handler
        ename = ename.lower()
        earr = self.__event_map.get(ename, None)
        if earr is None:
            earr = self.__event_map[ename] = []
        earr.remove(handler)

    def __trigger_event(self, etype, event):
        assert type(etype) is EventTypes
        assert type(event) is Event
        earr = self.__event_map[etype]
        for handler in earr:
            handler(event)

    def _get_trigger(self):
        if self.__get_trigger_count == 0:
            return self.__trigger_event
        raise LogicalError("_get_trigger can only be used by site itself. Do not try to get it after site got it."
                        "We are protecting it.")

    def trigger_custom_event(self, ename, event):
        assert ename is None
        assert type(event) is Event
        ename = ename.lower()
        earr = self.__custom_event_map[ename]
        if earr is None:
            return
        for handler in earr:
            handler(event)
        event.mark_complete()

    def handler_exists(self, etype, handler):
        assert type(etype) is EventTypes
        assert type(handler) is Handler
        earr = self.__event_map[etype]
        if handler in earr:
            return True
        else:
            return False

    def custom_handler_exists(self, ename, handler):
        assert ename is str
        assert type(handler) is Handler
        earr = self.__custom_event_map.get(ename, None)
        if earr is None:
            return False
        if handler in earr:
            return True
        else:
            return False




import re
from synamic.exceptions import SynamicModelParsingError, get_source_snippet_from_text
_key_pattern = re.compile(r'^[a-z0-9_]+(\.[a-z0-9_]+)*$', re.I)


class ModelField:
    def __init__(self, key, converter_name, required, unique):
        self.__key = key
        self.__converter_name = converter_name
        self.__required = required
        self.__unique = unique
        self.__converter = None

    def clone(self):
        return self.__class__(self.__key, self.__converter_name, self.__required, self.__unique)

    def set_converter(self, converter):
        # TODO: Importing ConverterCallable is impossible as this module is imported from type system.
        # Fix it or leave it.
        # assert isinstance(converter, ConverterCallable)
        assert self.__converter is None
        self.__converter = converter

    @property
    def key(self):
        return self.__key

    @property
    def converter_name(self):
        return self.__converter_name

    @property
    def converter(self):
        return self.__converter

    @property
    def required(self):
        return self.__required

    @property
    def unique(self):
        return self.__unique


class Model(dict):
    def __init__(self, model_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__model_name = model_name
        self.__args = tuple(args)
        self.__kwargs = kwargs.copy()

    def clone(self):
        another = self.__class__(self.__model_name, *self.__args, **self.__kwargs)
        for key, value in self.items():
            another[key] = value.clone()
        return another

    @property
    def model_name(self):
        return self.__model_name

    @property
    def body_field(self):
        return self.get('__body__', None)

    def new(self, *others):
        another = self.clone()
        for other in others:
            assert type(other) is self.__class__
            another.update(other.clone())
        return another


class ModelParser:
    @classmethod
    def empty_model(cls, model_name):
        return cls.parse(model_name, '')

    @staticmethod
    def parse(model_name, model_text):
        model = Model(model_name)
        lines = model_text.splitlines()
        line_no = 0
        for line in lines:
            line_no += 1
            line = line.strip()
            if line == '':
                # skip empty lines
                continue
            if line.startswith(('//', '#')):
                # skip comments
                continue
            else:
                # process
                key_value = line.split(':', 1)
                if len(key_value) == 1:
                    source_snippet = get_source_snippet_from_text(model_text, line_no, limit=10)
                    raise SynamicModelParsingError(f'Parsing error of model {model_name} at line {line_no}\n'
                                                   f'Details:\n{source_snippet}')
                else:
                    key, value = key_value
                    key = key.strip()
                    value = value.strip()

                    if not _key_pattern.match(key):
                        source_snippet = get_source_snippet_from_text(model_text, line_no, limit=10)
                        raise SynamicModelParsingError(f'Key {key} in model {model_name} did not match with key '
                                                       f'pattern at line no {line_no}\n'
                                                       f'Details:\n{source_snippet}')

                    parts = value.split('|')
                    parts = [part.strip() for part in parts]

                    converter = parts[0]
                    required = False
                    unique = False
                    parts_rest = parts[1:]
                    for part in parts_rest:
                        part = part.lower()
                        if part == 'required':
                            required = True
                        if part == 'unique':
                            unique = True
                    model_field = ModelField(key, converter, required, unique)
                    model[key] = model_field

        return model

